import re
import struct

from .portrequest import generate_port_message


class PortCommand:
    def __init__(
        self,
        module,
        name,
        byteid=None,
        receivetype=None,
        sendtype=None,
        receivefunction=None,
        sendfunction=None,
        arduino_code=None,
    ):
        global FIRSTFREEBYTEID
        self.module = module

        self.sendlength = struct.calcsize(sendtype) if sendtype is not None else 0
        self.receivelength = (
            struct.calcsize(receivetype) if receivetype is not None else 0
        )

        self.sendtype = sendtype
        self.receivetype = receivetype

        if sendfunction is None:
            sendfunction = self.defaultsendfunction

        self.sendfunction = sendfunction
        self.receivefunction = receivefunction
        self.name = re.sub(r"\s+", "", name, flags=re.UNICODE)
        if byteid == None:
            byteid = module.firstfreebyteid
        self.byteid = byteid
        self.arduino_code = ""
        self.set_arduino_code(arduino_code)

    def set_arduino_code(self, arduino_code):
        if arduino_code is not None:
            self.arduino_code = arduino_code.replace(
                "{BYTEID}", str(self.byteid)
            ).replace("{NAME}", str(self.name))
        else:
            self.arduino_code = ""

    def defaultsendfunction(self, numericaldata=None):
        if numericaldata is None:
            data = bytearray()
        else:
            data = struct.pack(self.sendtype, numericaldata)
        self.module.serialport.write(
            bytearray(generate_port_message(self.byteid, self.sendlength, *data))
        )

    def receive(self, bytearray):
        #print(self.receivetype,bytearray)
        self.receivefunction(self.module,struct.unpack(self.receivetype, bytearray)[0])



import collections

arduino_var_to_struc = collections.namedtuple('arduino_var_to_struc',
                                              'struct minimum maximum arduino_setter arduino_getter python_type html_input default_value byte_size')
arduino_var_to_struc_available = {
    'uint8_t': arduino_var_to_struc(struct="B", minimum=0, maximum=2 ** 8 - 1, arduino_setter="{{ardvar_name}}=data[0];",python_type=int, html_input='number',default_value=0,arduino_getter='write_data({{ardvar_name}},{BYTEID});', byte_size=1),
    'uint16_t': arduino_var_to_struc(struct="I", minimum=0, maximum=2 ** 16 - 1, arduino_setter="uint16_t temp;memcpy(&temp,data,2);{{ardvar_name}}=temp;",python_type=int, html_input='number',default_value=0,arduino_getter='write_data({{ardvar_name}},{BYTEID});', byte_size = 2),
    'uint32_t': arduino_var_to_struc(struct="L", minimum=0, maximum=2 ** 32 - 1, arduino_setter="uint32_t temp;memcpy(&temp,data,4);{{ardvar_name}}=temp;",python_type=int, html_input='number',default_value=0,arduino_getter='write_data({{ardvar_name}},{BYTEID});', byte_size = 4),
    'uint64_t': arduino_var_to_struc(struct="Q", minimum=0, maximum=2 ** 64 - 1, arduino_setter="uint64_t temp;memcpy(&temp,data,8);{{ardvar_name}}=temp;",python_type=int, html_input='number',default_value=0,arduino_getter='write_data({{ardvar_name}},{BYTEID});', byte_size = 8),
}


class StrucTypeNotFoundException(Exception): pass


class arduio_variable:
    def __init__(self, name, getter=None, setter=None, type='uint8_t', eeprom=False, default=None, sendtype=None,
                 receivetype=None, minimum=None, maximum=None, is_data_point=False,
                 python_type=None,html_input=None,arduino_getter=None,arduino_setter=None,byte_size=None,save=True):
        self.save = save
        self.is_data_point = is_data_point

        self.name = name
        self.eeprom = eeprom
        self.type = type
        try:
            ard_var=arduino_var_to_struc_available.get(type)
            self.receivetype = ard_var.struct if receivetype is None else receivetype
            self.sendtype = ard_var.struct if sendtype is None else sendtype
            self.maximum = ard_var.maximum if maximum is None else maximum
            self.minimum = ard_var.minimum if minimum is None else minimum

            self.byte_size = ard_var.byte_size if byte_size is None else byte_size
            
            self.arduino_setter = None if arduino_setter is False else (ard_var.arduino_setter.replace("{{ardvar_name}}",self.name) if arduino_setter is None else arduino_setter)
            self.arduino_getter = None if arduino_getter is False else (ard_var.arduino_getter.replace("{{ardvar_name}}",self.name) if arduino_getter is None else arduino_getter)

            self.python_type = arduino_var_to_struc_available.get(
                type).python_type if python_type is None else python_type

            self.html_input = '<input type='+ard_var.html_input+' min="'+str(self.minimum)+'" max="'+str(self.maximum)+'" value="{{value}}"'+(' readonly' if self.arduino_setter is None else '')+'>' if html_input is None else html_input
            self.default = ard_var.default_value if default is None else default
        except AttributeError:
            raise StrucTypeNotFoundException('Struct equivalent not found for ' + str(type) + ' please define manually')

#        if eeprom:
 #           self.arduino_setter=self.arduino_setter+""

        self.value = self.default
        self.setter = None if setter is False else (self.default_setter if setter is None else setter)
        self.getter = None if setter is False else (self.default_getter if getter is None else getter)

    def default_getter(self, instance):
        return self.value

    def default_setter(self, instance, data, send_to_board=True):
        if data is None: return
        data = self.python_type(data)
        if data < self.minimum: data = self.minimum
        if data > self.maximum: data = self.maximum
        if self.arduino_setter is not None:
            if send_to_board:
                instance.get_portcommand_by_name("set_"+self.name).sendfunction(data)
        self.value = data
        if self.is_data_point:
            instance.datapoint(self.name,data)

    def __set__(self, instance, value):
        self.setter(instance, value)

    def __get__(self, instance, owner):
        return self.getter(instance)

    def set_without_sending_to_board(self, instance, data):
        self.setter(instance, data, send_to_board=False)



from django import template

#if manage.py is called directly
if len(__name__.split(".")) == 2:
    from controll_server.settings import CONFIG
else:
    from ..controll_server.settings import CONFIG



register = template.Library()


@register.filter(name="pub_dict")
def get_from_public_dict(args, default):
    path = ["public"] + args.strip().split(",")
    return CONFIG.get(*path, default=default)



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys
import logging

logger = logging.getLogger("controll_server")
CONFIG = None

def run(*args):
    try:
        if __name__ == "__main__":
            from controll_server import settings
        else:
            from .controll_server import settings
        os.environ.setdefault(
            "DJANGO_SETTINGS_MODULE",settings.__name__
        )
        try:
            from django.core.management import execute_from_command_line
        except ImportError as exc:
            raise ImportError(
                "Couldn't import Django. Are you sure it's installed and "
                "available on your PYTHONPATH environment variable? Did you "
                "forget to activate a virtual environment?"
            ) from exc
        execute_from_command_line(args)
    except Exception as e:
        logger.exception(e)
        raise e


if __name__ == "__main__":
    run(*sys.argv)



import logging

from filter_dict import filter_dict


class TargetNotFoundException(Exception):
    pass


class PythonCommunicator():
    def __init__(self,logger=None):
        self._registered_nodes = dict()
        self.logger = logger if logger is not None else logging.getLogger("PythonCommunicator")


    def get_node(self, name):
        return self._registered_nodes.get(name)

    def add_node(self, name, target):
        self._registered_nodes[name] = target

    def cmd_out(self, cmd, targets=None, **kwargs):
        if targets is None:
            targets=list(self._registered_nodes.keys())

        for target in targets:
            node = self.get_node(target)
            if node is not None:
                if isinstance(node, PythonCommunicator):
                    node.cmd_in(cmd, **kwargs)
                else:
                    try:
                        method_to_call = getattr(node, cmd)
                        method_to_call(**filter_dict(kwargs, method_to_call))
                    except Exception as e:
                        #self.logger.exception(e)
                        pass
            else:
                raise TargetNotFoundException("target: "+str(target)+" not found in nodes ("+','.join(self._registered_nodes)+")")

    def cmd_in(self, cmd, **kwargs):
        print(cmd, kwargs)




"""digitalization_server URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.urls import path, include

#
#if manage.py is called directly
from multi_purpose_arduino_controller.controll_server.manage import logger

if len(__name__.split(".")) == 2:
    from templatetags.installed_apps import get_apps
else:
    from ..templatetags.installed_apps import get_apps
from . import views

urlpatterns = [path("admin/", admin.site.urls), path("", views.index, name="index")]


for app in get_apps():
    try:
        if hasattr(app, "baseurl"):
            print(app.label)
            urlpatterns.append(
                path(
                    app.baseurl + "/",
                    include(
                        ("%s.urls" % app.module_path, app.label), namespace=app.label
                    ),
                )
            )
            print(app.label)

    except ModuleNotFoundError as e:
        logger.exception(e)
        pass


import logging
import os
import tempfile
import time

from json_dict import JsonDict

from django_websocket_server.websocket_server import SocketServer
from multi_purpose_arduino_controller.python_communicator import PythonCommunicator, TargetNotFoundException
from .datalogger import DataLogger


class ArduinoControllerAPI:
    def __init__(
            self,
            name="arduinocontrollerapi",
            python_communicator: PythonCommunicator = None,
            websocket_server: SocketServer = None,
            config=None,
            datalogger: DataLogger = None,
            logger=None,
            data_dir=None,
    ):
        self._python_communicator = None
        self.boarddata = {}
        self.websocket_server = None
        if websocket_server is not None: self.set_websocket_server(websocket_server)
        self.name = name
        self.data = {}
        self.dataupdate = 1
        self.lastupdate = 0
        self.lastsend = dict()
        self.connected_ports = set()
        self.ignored_ports = set()
        self.available_ports = set()
        self.identified_ports = set()

        self.logger = logging.getLogger(self.name) if logger is None else - logger

        self.data_dir = os.path.join(tempfile.gettempdir(), self.name + "_API") if data_dir is None else data_dir

        os.makedirs(self.data_dir, exist_ok=True)

        self.config = JsonDict(file=os.path.join(self.data_dir, "config.json"),
                               createfile=True) if config is None else config
        self.config.autosave = True

        self.sensor_ports = set([])

        self.set_communicator(PythonCommunicator() if python_communicator is None else python_communicator)

        self.datalogger = DataLogger(autosave_path=self.data_dir) if datalogger is None else datalogger

        self.running = False
        self.run_thread = None

    def set_communicator(self, communicator: PythonCommunicator):
        self._python_communicator = communicator
        self._python_communicator.add_node(self.name, self)

    def get_communicator(self):
        return self._python_communicator

    python_communicator = property(get_communicator,set_communicator)

    def ask_for_ports(self):
        try:
            self._python_communicator.cmd_out(
                cmd="get_ports",
                sender=self.name,
                targets=["serialreader"],
                data_target=self.name,
            )
        except TargetNotFoundException:
            pass

    def set_ports(self, connected_ports=None, ignored_ports=None, available_ports=None,identified_ports=None):
        self.connected_ports = set([] if connected_ports is None else connected_ports)
        self.ignored_ports = set([] if ignored_ports is None else ignored_ports)
        self.available_ports = set([] if available_ports is None  else available_ports)
        self.identified_ports = set([] if identified_ports is None  else identified_ports)
        for port in self.connected_ports.difference(self.sensor_ports): self.add_sensor_port(port)
        for port in self.sensor_ports.difference(self.connected_ports): self.add_sensor_port(port)

    def remove_sensor_port(self, port):
        self.logger.info("remove sensor port" + str(port))
        self.sensor_ports.remove(port)

    def add_sensor_port(self, port=None):
        time.sleep(1)
        if port is not None:
            self.logger.info("add sensor port" + str(port))
            self.sensor_ports.add(port)
            self._python_communicator.cmd_out(
                cmd="add_data_target", targets=[port], data_target=self.name
            )

    def data_update_time(self, data_update_time=None):
        if data_update_time is not None:
            self.dataupdate = data_update_time
            self.lastupdate = data_update_time.time() - self.dataupdate

    def datapoint(self, key=None, x=None, y=None,**kwargs):
        if key is None or x is None or y is None:
            return

        self.datalogger.add_datapoint(key, x, y)

        # self.data[message["data"]["key"]].append([message["data"]["x"], message["data"]["y"], message["data"]["t"]])

        t = time.time()
        if t - self.lastupdate > self.dataupdate:
            self.lastupdate = t

            for key, values in self.datalogger.get_last_valid_values().items():
                if key not in self.lastsend:
                    self.lastsend[key] = ("x", "y")
                if (
                        self.lastsend[key][0] != values[0]
                        or self.lastsend[key][1] != values[1]
                ):
                    self.lastsend[key] = (values[0], values[1])
                    #self._python_communicator.cmd_out(
                    #    cmd="datapoint",
                    #    key=key,
                    #    x=values[0],
                    #    y=values[1],
                    #    targets=["websocket"],
                    #)

            # for key, dates in self.data.items():
            #   self.communicator.cmd_out(
            #          sender=self.name,
            #         x=dates[-1][0],
            #        y=dates[-1][1],
            #       key=key,
            #      target="gui",
            #     t=dates[-1][2],
            #    as_string=True,
            # )

    def get_data(self, data_target=None):
        if data_target is None:
            return
        self._python_communicator.cmd_out(
            sender=self.name,
            cmd="set_data",
            target=data_target,
            as_string=True,
            data=self.data,
        )

    def boardupdate(self, boarddata=None):
        if boarddata is None:
            return
        self.boarddata[boarddata.get('port')] = boarddata
        self.logger.info("boardupdate:" + str(boarddata))

    def get_board_attribute(self, port, attribute):
        if port not in self.boarddata: return
        return self.boarddata[port].get(attribute, None)

    def port_opened(self,port):
        pass

    def port_closed(self,port):
        if port in self.sensor_ports:
            self.sensor_ports.remove(port)

    def port_identified(self,port):
        pass

    def set_websocket_server(self, websocket_server):
        self.websocket_server = websocket_server




import time
from collections import OrderedDict

import numpy as np

from multi_purpose_arduino_controller.arduino_controller.arduino_variable import arduio_variable
from multi_purpose_arduino_controller.arduino_controller.basicboard import arduino_data
from multi_purpose_arduino_controller.arduino_controller.basicboard.ino_creator import InoCreator
from multi_purpose_arduino_controller.arduino_controller.basicboard.pin import Pin
from multi_purpose_arduino_controller.arduino_controller.portcommand import PortCommand
import inspect

MAXATTEMPTS = 3
IDENTIFYTIME = 2


# noinspection PyBroadException
class ArduinoBasicBoard:
    FIRMWARE = 0

    FIRSTFREEBYTEID = 0

    def get_first_free_byteid(self):
        ffbid = self.FIRSTFREEBYTEID
        self.FIRSTFREEBYTEID += 1
        return ffbid

    firstfreebyteid = property(get_first_free_byteid)

    firmware = arduio_variable(name='firmware', type='uint64_t', arduino_setter=False,default=-1)
    datarate = arduio_variable(name='datarate', type='uint32_t',default=200,minimum=1,eeprom=True)


    def __init__(self):
        self.inocreator = InoCreator(self)
        self.inocreator.add_creator(arduino_data.create)
        self._pins = dict()
        self.save_attributes = OrderedDict()
        self.static_attributes = set()
        self.free_digital_pins = list(range(2, 12))
        self.name = None
        self.port = None
        self.lastdata = None
        self.updatetime = 2
        self.identify_attempts = 0

        self.save_attributes.update(
            {
                "firmware": "int+",
                "port": "string",
                "id": "int+",
                "name": "string",
                "updatetime": "double+",
                "pins": "int+0",
            }
        )

        self.static_attributes.update(["firmware", "id", "port"])
        self.identified = False
        self.id = None
        self.portcommands = []
        self.serialport = None
        self.add_port_command(
            PortCommand(
                module=self,
                name="identify",
                receivetype="Q",
                sendtype="?",
                receivefunction=ArduinoBasicBoard.receive_id,
                byteid=self.firstfreebyteid,
                arduino_code="identified=data[0];uint64_t id = get_id();write_data(id,{BYTEID});",
            )
        )

        for attr, ard_var in self.get_arduino_vars().items():
            if ard_var.arduino_setter is not None:
                self.add_port_command(
                    PortCommand(
                        module=self,
                        byteid=self.get_first_free_byteid(),
                        name="set_" + ard_var.name,
                        sendtype=ard_var.sendtype,
                        receivetype=None,
                        receivefunction=ard_var.set_without_sending_to_board,
                        arduino_code=ard_var.arduino_setter,
                    )
                )
            if ard_var.arduino_getter is not None:
                self.add_port_command(
                    PortCommand(
                        module=self,
                        byteid=self.get_first_free_byteid(),
                        name="get_" + ard_var.name,
                        sendtype=None,
                        receivetype=ard_var.receivetype,
                        receivefunction=ard_var.set_without_sending_to_board,
                        arduino_code=ard_var.arduino_getter,
                    )
                )

    def get_arduino_vars(self):
        ardvars = {}
        classes = inspect.getmro(self.__class__)
        for cls in reversed(classes):
            for attr, ard_var in cls.__dict__.items():
                if isinstance(ard_var, arduio_variable):
                    ardvars[attr] = ard_var
        return ardvars


    def set_serial_port(self, serialport):
        self.serialport = serialport
        self.port = serialport.port
        if self.name is None:
            self.name = self.port

    def add_pin(self, pinname, defaultposition, pintype=Pin.DIGITAL_OUT):
        portcommand = PortCommand(
            module=self,
            name=pinname,
            receivetype="B",
            sendtype="B",
            receivefunction=lambda data: (ArduinoBasicBoard.set_pin(pinname, data, to_board=False)),
            byteid=self.firstfreebyteid,
        )
        pin = Pin(
            name=pinname,
            defaultposition=defaultposition,
            portcommand=portcommand,
            pintype=pintype,
        )
        self.set_pin(pinname, pin, to_board=False)
        self.add_port_command(portcommand)

    def get_first_free_digitalpin(self, catch=True):
        fp = self.free_digital_pins[0]
        if catch:
            self.free_digital_pins.remove(fp)
        return fp

    def identify(self):
        from multi_purpose_arduino_controller.arduino_controller.serialport import BAUDRATES
        for b in set([self.serialport.baudrate] + list(BAUDRATES)):
            self.identify_attempts = 0
            self.serialport.logger.info(
                "intentify with baud " + str(b) + " and firmware " + str(self.FIRMWARE)
            )
            try:
                self.serialport.baudrate = b
                while self.id is None and self.identify_attempts < MAXATTEMPTS:
                    self.get_portcommand_by_name("identify").sendfunction(0)
                    self.identify_attempts += 1
                    time.sleep(IDENTIFYTIME)
                if self.id is not None:
                    self.identified = True
                    break
            except Exception as e:
                self.serialport.logger.exception(e)
                pass
        if not self.identified:
            return False

        self.identified = False
        self.identify_attempts = 0
        while self.firmware == -1 and self.identify_attempts < MAXATTEMPTS:
            self.get_portcommand_by_name("get_firmware").sendfunction()
            self.identify_attempts += 1
            time.sleep(IDENTIFYTIME)
        if self.firmware > -1:
            self.identified = True
        return self.identified

    #def specific_identification(self):
    #    self.identified = False
    #    self.identify_attempts = 0
    #    while self._datarate <= 0 and self.identify_attempts < MAXATTEMPTS:
    #        self.get_portcommand_by_name("datarate").sendfunction(0)
    #        self.identify_attempts += 1
    #        time.sleep(IDENTIFYTIME)
    #    if self._datarate > 0:
    #        self.identified = True
    #    if not self.identified:
    #        return False

     #   return self.identified

    def receive_from_port(self, cmd, data):
        self.serialport.logger.debug(
            "receive from port cmd: " + str(cmd) + " " + str([i for i in data])
        )
        portcommand = self.get_portcommand_by_cmd(cmd)
        if portcommand is not None:
            portcommand.receive(data)
        else:
            self.serialport.logger.debug("cmd " + str(cmd) + " not defined")

    def add_port_command(self, port_command):
        if (
                self.get_portcommand_by_cmd(port_command.byteid) is None
                and self.get_portcommand_by_name(port_command.name) is None
        ):
            self.portcommands.append(port_command)
        else:
            self.serialport.logger.error(
                "byteid of "
                + str(port_command)
                + " "
                + port_command.name
                + " already defined"
            )

    def get_portcommand_by_cmd(self, byteid):
        for p in self.portcommands:
            if p.byteid == byteid:
                return p
        return None

    def get_portcommand_by_name(self, command_name):
        for p in self.portcommands:
            if p.name == command_name:
                return p
        return None

    def receive_id(self, data):
        self.id = int(np.uint64(data))

    def datapoint(self, name, data):
        self.lastdata = data
        if self.identified:
            self.serialport.add_datapoint(self,key=str(self.id) + "_"+str(name), y=data, x=None)

    def restore(self, data):
        for key, value in data.items():
            if key not in self.static_attributes:
                if getattr(self, key, None) != value:
                    setattr(self, key, value)

        for attr, ard_var in self.get_arduino_vars().items():
            if ard_var.save and attr in data:
                setattr(self, attr, data[attr])

    def set_pins(self, pindict):
        for pin_name, pin in pindict.items():
            self.set_pin(pin_name, pin)

    def get_pins(self):
        return self._pins

    pins = property(get_pins, set_pins)

    def set_pin(self, pin_name, pin, to_board=True):
        if isinstance(pin, Pin):
            if self._pins.get(pin_name, None) == pin:
                return
            elif self._pins.get(pin_name, None) is not None:
                if self._pins[pin_name].position == pin.position:
                    self._pins[pin_name] = pin
                    return
            else:
                self._pins[pin_name] = pin
        else:
            if pin_name in self._pins:
                if self._pins[pin_name].position == pin:
                    return
                else:
                    self._pins[pin_name].position = pin
            else:
                return
        try:
            self.serialport.logger.info(
                "set Pin " + pin_name + " to " + str(self._pins[pin_name].position)
            )
        except Exception as e:
            pass
        if to_board:
            self._pins[pin_name].portcommand.sendfunction(pin)

    def get_pin(self, pin_name):
        return self._pins.get(pin_name, None)

    def serialize_attribute(self, value):
        try:
            return value.to_json()
        except Exception as e:
            pass

        if isinstance(value, dict):
            return {key: self.serialize_attribute(val) for key, val in value.items()}
        if isinstance(value, list):
            return [self.serialize_attribute(val) for val in value]
        return value

    def save(self):
        data = {}
        for attribute in self.save_attributes:
            val = getattr(self, attribute, None)
            val = self.serialize_attribute(val)
            data[attribute] = val
        for attr, ard_var in self.get_arduino_vars().items():
            if ard_var.save:
                data[attr] = ard_var.value
        return data

    def get_board(self):
        board = {
            'arduino_variables': {}
        }
        for attr, ard_var in self.get_arduino_vars().items():
            board['arduino_variables'][attr] = {
                'form': ard_var.html_input.replace("{{value}}", str(getattr(self, attr, '')))
            }
        return board



    def create_ino(self):
        import inspect
        import os
        self.firmware = self.FIRMWARE
        ino = self.inocreator.create()
        dir = os.path.dirname(inspect.getfile(self.__class__))
        name = os.path.basename(dir)
        with open(os.path.join(dir, name + ".ino"), "w+") as f:
            f.write(ino)


if __name__ == "__main__":
    ins = ArduinoBasicBoard()
    ins.create_ino()


import sys
import glob
import serial


def get_avalable_serial_ports(ignore=None):
    """ Lists serial port names
        :raises EnvironmentError:
            On unsupported or unknown platforms
        :returns:
            A list of the serial ports available on the system
    """
    if ignore is None:
        ignore = set()
    if sys.platform.startswith("win"):
        ports = ["COM%s" % (i + 1) for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob("/dev/tty[A-Za-z]*")
    elif sys.platform.startswith("darwin"):
        ports = glob.glob("/dev/tty.*")
    else:
        raise EnvironmentError("Unsupported platform")

    ports = set(ports)

    result = []
    for port in ports.difference(ignore):
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return set(result), ignore.intersection(ports)


if __name__ == "__main__":
    print(get_avalable_serial_ports())



from . import defaultcodes


class Pin:
    DIGITAL_OUT = 0

    def __init__(self, name, defaultposition, portcommand, pintype=0):
        self.name = name
        self.portcommand = portcommand
        self.position = defaultposition
        self.pintype = pintype

        if pintype == self.DIGITAL_OUT:
            self.portcommand.set_arduino_code(
                defaultcodes.DEFAULTCODES["setpin_output"]
            )
        else:
            self.portcommand.set_arduino_code(
                defaultcodes.DEFAULTCODES["setpin_output"]
            )

    def arduinoMode(self):
        if self.pintype == self.DIGITAL_OUT:
            return "OUTPUT"
        else:
            return "OUTPUT"

    def to_json(self):
        return self.position



import os
import sys
import threading
import time

import pandas as pd
import numpy as np


class DataLogger:
    def __init__(self, rolling_data=False, rolling_size=1000,autosave_path=None,auosave_name="data",autosave=False):
        self._rolling_size = rolling_size
        self._max_save_size = None
        self._min_save_size = None
        self._min_save_time = None
        self._max_save_time = None
        self._save_check_time = None
        self._savename = None
        self._save_path = None
        self._autosave = False
        self._last_save_time = 0.0
        self.starttime = 0.0
        self._min_save_size = 50
        self.autosave_files = set()
        self.__autosave_thread: threading.Thread = None

        self._df = pd.DataFrame()
        self._df_rolling = None
        if rolling_data:
            self._df_rolling = pd.DataFrame()

        if autosave_path is not None and autosave:
            self.start_autosave(autosave_path,auosave_name)

    def start_autosave(
        self,
        path,
        savename,
        save_check_time=20,
        max_save_time=60 * 10,
        max_save_size=5 * 10 ** 6,
        min_save_size=50,
    ):
        self.stop_autosave()
        self.autosave_files=set()
        self._df = pd.DataFrame()
        self._min_save_size = min_save_size
        self._max_save_size = max_save_size
        self._max_save_time = max_save_time
        self._save_check_time = save_check_time
        self._savename = savename
        self._save_path = path
        self.starttime = time.time()
        self._autosave = True
        self.__autosave_thread = threading.Thread(target=self._autosaver)
        self.__autosave_thread.start()

    def stop_autosave(self, mergedata=False):
        if self.__autosave_thread is not None:
            self._autosave = False
            try:
                self.__autosave_thread.join()
            except:
                pass
            self.__autosave_thread = None
            s = sys.getsizeof(self._df)
            if s > self._min_save_size:
                self._do_autosave()

        if mergedata:
            li = []
            fullname = os.path.join(self._save_path, self._savename +"_" + str(int(self.starttime))+"_" + str(int(self._last_save_time)) + ".csv")
            for file in sorted(list(self.autosave_files)):
                df = pd.read_csv(file, index_col=0)
                li.append(df)

            try:
                df = pd.DataFrame(pd.concat(li, axis=0))
                df.to_csv(fullname)
                for file in self.autosave_files:
                    try:
                        os.remove(file)
                    except:
                        pass
            except ValueError:
                return None
            self.autosave_files = set()
            return fullname
        return self.autosave_files

    def _do_autosave(self):
        self._last_save_time = time.time()
        savedf = pd.DataFrame()
        savedf, self._df = self._df, savedf
        savename=self._savename + "_" + str(int(self._last_save_time)) + ".csv"
        savefile = os.path.join(self._save_path,savename,)
        self.autosave_files.add(savefile)
        savedf.to_csv(savefile)

    def _autosaver(self):
        while self._autosave:
            dt = time.time() - self._last_save_time
            s = sys.getsizeof(self._df)
            if dt > self._max_save_time or s > self._max_save_size:
                if s > self.\
                        _min_save_size:
                    self._do_autosave()
            time.sleep(1)
        self.stop_autosave()

    def add_datapoint(self, key, x, y):
        if key not in self._df.columns:
            self._df[key] = np.nan

        if x not in self._df.index:
            self._df.loc[x] = np.nan
        self._df[key][x] = y

        if self._df_rolling is not None:
            if key not in self._df_rolling.columns:
                self._df_rolling[key] = np.nan
            if x not in self._df_rolling.index:
                self._df_rolling.loc[x] = np.nan
            self._df_rolling[key][x] = y

            if self._df_rolling.shape[0] > self._rolling_size:
                self._df_rolling = self._df_rolling.iloc[-self._rolling_size :]

    def get_last_valid_values(self):
        last_valid = {}
        for key in self._df.columns:
            x = self._df[key].last_valid_index()
            if x is not None:
                y = self._df[key][x]
                last_valid[key] = (x, y)
        return last_valid



import struct

STARTBYTE = bytes([2])
STARTBYTEPOSITION = 0
COMMANDBYTEPOSITION = 1
LENBYTEPOSITION = 2
DATABYTEPOSITION = 3


def generate_checksum(array):
    try:
        array = [ord(i) for i in array]
    except:
        pass
    sum1 = 0
    sum2 = 0
    for i in array:
        sum1 = (sum1 + i) % 255
        sum2 = (sum2 + sum1) % 255
    return sum2 * 256 + sum1


def generate_port_message(cmd, datalength, *args):
    assert len(args) >= datalength
    return generate_request(cmd, args[:datalength])


def generate_request(command, data):
    a = [2, command, len(data), *data]
    a.extend(struct.pack(">H", generate_checksum(a)))
    return bytearray(a)


def validate_buffer(port):
    try:
        firststart = port.readbuffer.index(STARTBYTE)
    except ValueError:
        firststart = 0
        port.readbuffer = []

    bufferlength = len(port.readbuffer[firststart:])
    if bufferlength >= DATABYTEPOSITION + 2:
        datalength = ord(port.readbuffer[firststart + LENBYTEPOSITION])
        if bufferlength >= DATABYTEPOSITION + datalength + 2:
            databuffer = port.readbuffer[
                firststart
                + DATABYTEPOSITION : firststart
                + DATABYTEPOSITION
                + datalength
            ]
            checksum, = struct.unpack(
                ">H",
                b"".join(
                    port.readbuffer[
                        firststart
                        + DATABYTEPOSITION
                        + datalength : firststart
                        + DATABYTEPOSITION
                        + datalength
                        + 2
                    ]
                ),
            )
            if checksum == generate_checksum(
                port.readbuffer[firststart : firststart + DATABYTEPOSITION + datalength]
            ):
                port.board.receive_from_port(
                    cmd=ord(port.readbuffer[firststart + COMMANDBYTEPOSITION]),
                    data=b"".join(databuffer),
                )
            port.readbuffer = port.readbuffer[
                firststart + DATABYTEPOSITION + datalength + 2 :
            ]



class InoCreator:
    def __init__(self, board):
        self.creators = []
        self.board = board
        self.data = dict(
            definitions={},
            global_vars={},
            includes=set([]),
            functions={},
            setup="",
            loop="",
            dataloop="",
        )

    def add(
        self,
        definitions=None,
        global_vars=None,
        includes=None,
        functions=None,
        setup=None,
        loop=None,
        dataloop=None,
    ):
        if definitions is not None:
            self.add_definitions(definitions)
        if global_vars is not None:
            self.add_global_vars(global_vars)
        if includes is not None:
            self.add_includes(includes)
        if functions is not None:
            self.add_functions(functions)
        if setup is not None:
            self.add_setup(setup)
        if loop is not None:
            self.add_loop(loop)
        if dataloop is not None:
            self.add_dataloop(dataloop)

    def add_definitions(self, definitions):
        for name, value in definitions.items():
            self.add_definition(name, value)

    def add_definition(self, name, value):
        if name in self.data["definitions"]:
            raise ValueError(name + " already defined")
        self.data["definitions"][name] = value

    def add_global_vars(self, vars):
        for name, value in vars.items():
            self.add_global_var(name, value)

    def add_global_var(self, name, value):
        if name in self.data["global_vars"]:
            raise ValueError(name + " already defined")
        self.data["global_vars"][name] = value

    def add_includes(self, includes):
        self.data["includes"].update(includes)

    def add_functions(self, functions):
        for name, value in functions.items():
            self.add_function(name, value)

    def add_function(self, name, value):
        if name in self.data["functions"]:
            raise ValueError(name + " already defined")
        self.data["functions"][name] = value

    def add_setup(self, setup):
        self.data["setup"] = self.data["setup"] + setup

    def add_loop(self, loop):
        self.data["loop"] = self.data["loop"] + loop

    def add_dataloop(self, dataloop):
        self.data["dataloop"] = self.data["dataloop"] + dataloop

    def create(self):
        for creator in self.creators:
            self.add(**creator(self.board))
        text = ""
        for name, value in self.data["definitions"].items():
            text += "#define " + name + " " + str(value) + "\n"
        for inc in self.data["includes"]:
            text += "#include " + inc + "\n"
        for name, value in self.data["global_vars"].items():
            text += (
                value[0]
                + " "
                + name
                + (" = " + str(value[1]) if value[1] is not None else "")
                + ";\n"
            )
        for name, value in self.data["functions"].items():
            text += (
                value[0]
                + " "
                + name
                + "("
                + ", ".join([arg[0] + " " + arg[1] for arg in value[1]])
                + "){\n"
                + value[2]
                + "}\n"
            )

        text += "\nvoid dataloop(){\n" + self.data["dataloop"] + "\n}\n"
        text += "\nvoid loop(){\n" + self.data["loop"] + "\n}\n"
        text += "\nvoid setup(){\n" + self.data["setup"] + "\n}\n"
        return text

    def add_creator(self, creator):
        self.creators.append(creator)



import logging
import threading
import time

from json_dict import JsonDict

from multi_purpose_arduino_controller import python_communicator
from ..datalogger import DataLogger
from ..serialreader import serialdetector

AUTOCHECKPORTS = True
PORTCHECKTIME = 2


class SerialReader:
    def __init__(
            self,
            config: JsonDict = None,
            communicator: python_communicator.PythonCommunicator = None,
            logger=None,
            autocheckports=AUTOCHECKPORTS,
            port_check_time=PORTCHECKTIME,
            permanently_ignored_ports=None,
            datalogger: DataLogger = None,
            start_in_background=False
    ):

        self.ignored_ports = set()
        self.deadports = set()
        self.connected_ports = set()
        self.available_ports = set()
        self.identified_ports = set()

        self.datalogger = DataLogger() if datalogger is None else datalogger

        self.config = JsonDict("portdata.json") if config is None else config

        if permanently_ignored_ports is None:
            permanently_ignored_ports = []

        self.port_check_time = port_check_time
        self.autocheckports = autocheckports
        self.permanently_ignored_ports = set(permanently_ignored_ports)

        self.set_communicator(python_communicator.PythonCommunicator() if communicator is None else communicator)


        self.logger = logging.getLogger("SerialReader") if logger is None else logger

        if start_in_background:
            self.run_in_background()

    def reactivate_port(self,port=None):
        if port is None:
            return
        try:
            self.ignored_ports.remove(port)
        except:
            pass

        try:
            self.deadports.remove(port)
        except:
            pass

    def deactivate_port(self,port=None):
        if port is None:
            return
        self.ignored_ports.add(port)
        self._communicator.cmd_out(targets=[port], cmd="stop_read")

    def run_in_background(self):
        threading.Thread(target=self.read_forever).start()

    def set_communicator(self, communicator: python_communicator.PythonCommunicator):
        self._communicator = communicator
        self._communicator.add_node("serialreader", self)

    def get_communicator(self):
        return self._communicator

    communicator = property(get_communicator, set_communicator)

    def sendports(self, data_target=None):
        try:
            self._communicator.cmd_out(
                cmd="set_ports",
                sender="serialreader",
                targets=(None if data_target is None else (data_target if isinstance(data_target,list) else [data_target])),
                available_ports=list(self.available_ports),
                ignored_ports=list(self.ignored_ports | self.permanently_ignored_ports),
                connected_ports=[sp.port for sp in self.connected_ports],
                identified_ports=[sp.port for sp in self.identified_ports],
            )
        except python_communicator.TargetNotFoundException:
            pass

    get_ports = sendports

    def read_forever(self):
        while 1:
            if self.autocheckports:
                self.available_ports, self.ignored_ports = serialdetector.get_avalable_serial_ports(
                    ignore=self.ignored_ports | self.permanently_ignored_ports
                )
                self.deadports = self.available_ports.intersection(self.deadports)
                newports = self.available_ports - (
                        self.ignored_ports | self.deadports | self.permanently_ignored_ports
                )
                self.logger.debug(
                    "available Ports: "
                    + str(self.available_ports)
                    + "; new Ports: "
                    + str(newports)
                    + "; ignored Ports: "
                    + str(self.ignored_ports | self.permanently_ignored_ports)
                )

                self.sendports()
                for port in newports.copy():
                    try:
                        self.open_port(port)
                    except Exception as e:
                        self.logger.exception(e)
                        pass
            time.sleep(self.port_check_time)

    def open_port(self, port):
        try:
            self.available_ports.remove(port)
        except:
            pass
        try:
            self.ignored_ports.remove(port)
        except:
            pass
        try:
            self.permanently_ignored_ports.remove(port)
        except:
            pass

        from ..serialport import SerialPort
        t = threading.Thread(
            target=SerialPort,
            kwargs={
                **{
                    "communicator": self._communicator,
                    "serialreader": self,
                    "config": self.config,
                    "port": port,
                    "baudrate": self.config.get("portdata", port, "baud", default=9600),
                },
                # **kwargs,
            },
        )
        t.start()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    sr = SerialReader()

    sr.read_forever()



from django.apps import apps


APPS = None
CONTEXTAPPS = None


def reset():
    global APPS, CONTEXTAPPS
    APPS = None
    CONTEXTAPPS = None


def get_apps():
    global APPS
    if APPS is None:
        APPS = [app for app in apps.get_app_configs()]
    return APPS


def get_apps_context(request):
    global CONTEXTAPPS
    if CONTEXTAPPS is None:
        CONTEXTAPPS = []
        for app in get_apps():
            if hasattr(app, "to_context"):
                if app.to_context:
                    CONTEXTAPPS.append(
                        {
                            "indexurl": (
                                app.indexurl
                                if hasattr(app, "indexurl")
                                else app.label + ":index"
                            ),
                            "in_nav_bar": (
                                app.in_nav_bar if hasattr(app, "in_nav_bar") else False
                            ),
                            "navbarhtml":(app.in_nav_bar_html if hasattr(app, "in_nav_bar_html") else False),
                            "verbose_name": app.verbose_name,
                        }
                    )
    apps_in_nav = False
    for app in CONTEXTAPPS:
        if app["in_nav_bar"]:
            apps_in_nav = True
            break

    return {"apps": CONTEXTAPPS,
            'apps_in_nav':apps_in_nav
            }



from ..arduino_variable import arduio_variable

FIRSTFREEEEPROM = 8+2 #id + idcs

def definitions(self):
    from ..portrequest import DATABYTEPOSITION

    return {
        "STARTANALOG": 0,
        "ENDANALOG": 100,
        "STARTBYTE": 2,
        "STARTBYTEPOSITION": 0,
        "COMMANDBYTEPOSITION": 1,
        "LENBYTEPOSITION": 2,
        "DATABYTEPOSITION": 3,
        "MAXFUNCTIONS": str(len(self.portcommands)),
        "SERIALARRAYSIZE": str(
            DATABYTEPOSITION
            + max(
                *[
                    max(portcommand.receivelength, portcommand.sendlength)
                    for portcommand in self.portcommands
                ]
            )
            + 2
        ),
        "BAUD": 9600,
     #   "DATARATE": 200,
    }


def global_vars(self):
    print({ard_var.name: [ard_var.type, ard_var.value] for attr,ard_var in self.get_arduino_vars().items()},)
    return {
        **{name: ["uint8_t", str(pin.position)] for name, pin in self.pins.items()},
        **{ard_var.name: [ard_var.type, ard_var.value] for attr,ard_var in self.get_arduino_vars().items()},
        "writedata[SERIALARRAYSIZE]": ["uint8_t", None],
        "serialread[SERIALARRAYSIZE]": ["uint8_t", None],
        "serialreadpos": ["uint8_t", 0],
        "commandlength": ["uint8_t", 0],
        "cmds[MAXFUNCTIONS ]": ["uint8_t", None],
        "cmd_length[MAXFUNCTIONS]": ["uint8_t", None],
        "(*cmd_calls[MAXFUNCTIONS])(uint8_t* data, uint8_t s)": ["void", None],
        "lastdata": ["uint32_t", 0],
        "ct": ["uint32_t", None],
   #     "datarate": ["uint32_t", "DATARATE"],
        "c": ["uint8_t", None],
        "identified": ["bool", "false"],
    }


def includes(self):
    return ["<EEPROM.h>"]


def functions(self):
    return {
        "generate_checksum": [
            "uint16_t",
            [("uint8_t*", "data"), ("int", "count")],
            "uint16_t sum1 = 0;\n"
            "uint16_t sum2 = 0;\n"
            "for (int index = 0; index < count; ++index ) {\n"
            "sum1 = (sum1 + data[index]) % 255;\n"
            "sum2 = (sum2 + sum1) % 255;\n"
            "}\n"
            "return (sum2 << 8) | sum1;\n",
        ],
        "write_data_array": [
            "void",
            [("uint8_t*", "data"), ("uint8_t", "cmd"), ("uint8_t", "len")],
            "writedata[STARTBYTEPOSITION] = STARTBYTE;\n"
            "writedata[COMMANDBYTEPOSITION] = cmd;\n"
            "writedata[LENBYTEPOSITION] = len;\n"
            "for (uint8_t i = 0; i < len; i++) {\n"
            "writedata[DATABYTEPOSITION + i] = data[i];\n"
            "}"
            "uint16_t cs = generate_checksum(writedata, len + DATABYTEPOSITION);\n"
            "writedata[DATABYTEPOSITION + len] = cs >> 8;\n"
            "writedata[DATABYTEPOSITION + len + 1] = cs >> 0;\n"
            "Serial.write(writedata, len + DATABYTEPOSITION + 2);\n",
        ],
        "write_data": [
            "template< typename T> void",
            [("T", "data"), ("uint8_t", "cmd")],
            "uint8_t d[sizeof(T)];\n"
            "for (uint8_t i = 0;i<sizeof(T) ; i++) {\n"
            "d[i] = (uint8_t) (data >> (8 * i) & 0xff );\n"
            "}\n"
            "write_data_array(d, cmd, sizeof(T));\n",
        ],
        "get_id": [
            "uint64_t",
            [],
            "uint64_t id;\n" "EEPROM.get(0, id);\n" "return id;\n",
        ],
        "checkUUID": [
            "void",
            [],
            "uint64_t id = get_id();\n"
            "uint16_t cs = generate_checksum((uint8_t*)&id, sizeof(id));\n"
            "uint16_t cs2;\n"
            "EEPROM.get(sizeof(id), cs2);\n"
            "if (cs != cs2) {\n"
            "id = (uint64_t)("
            "(((uint64_t)random()) << 48) | "
            "(((uint64_t)random()) << 32) | "
            "(((uint64_t)random()) << 16) | "
            "(((uint64_t)random()))"
            ");\n"
            "EEPROM.put(0, id);\n"
            "EEPROM.put(sizeof(id), generate_checksum((uint8_t*)&id, sizeof(id)));\n"
            "}\n",
        ],
        "add_command": [
            "void",
            [
                ("uint8_t", "cmd"),
                ("uint8_t", "len"),
                ("void", "(*func)(uint8_t* data, uint8_t s)"),
            ],
            "for (uint8_t i = 0; i < MAXFUNCTIONS; i++ ) {\n"
            "if (cmds[i] == 255) {\n"
            "cmds[i] = cmd;\n"
            "cmd_length[i] = len;\n"
            "cmd_calls[i] = func;\n"
            "return;\n"
            "}\n"
            "}\n",
        ],
        "endread": [
            "void",
            [],
            "commandlength = 0;\n" "serialreadpos = STARTBYTEPOSITION;\n",
        ],
        "get_cmd_index": [
            "uint8_t",
            [("uint8_t", "cmd")],
            "for (uint8_t i = 0; i < MAXFUNCTIONS; i++ ) {\n"
            "if (cmds[i] == cmd) {\n"
            "return i;\n"
            "}\n"
            "}\n"
            ""
            "return 255;",
        ],
        "validate_serial_command": [
            "void",
            [],
            "if(generate_checksum(serialread, DATABYTEPOSITION + serialread[LENBYTEPOSITION]) == (uint16_t)(serialread[DATABYTEPOSITION + serialread[LENBYTEPOSITION]] << 8) + serialread[DATABYTEPOSITION + serialread[LENBYTEPOSITION]+1]){\n"
            "uint8_t cmd_index = get_cmd_index(serialread[COMMANDBYTEPOSITION]);\n"
            "if(cmd_index != 255){\n"
            "uint8_t data[serialread[LENBYTEPOSITION]];\n"
            "memcpy(data,&serialread[DATABYTEPOSITION],serialread[LENBYTEPOSITION]);\n"
            "cmd_calls[cmd_index](data,serialread[LENBYTEPOSITION]);\n"
            "}\n"
            "}\n",
        ],
        "readloop": [
            "uint64_t",
            [],
            "while(Serial.available() > 0) {\n"
            "c = Serial.read();\n"
            "serialread[serialreadpos] = c;\n"
            "if (serialreadpos == STARTBYTEPOSITION) {\n"
            "if (c == STARTBYTE) {\n"
            "} else {\n"
            "endread();\n"
            "continue;\n"
            "}\n"
            "}\n"
            "else {\n"
            "if (serialreadpos == LENBYTEPOSITION) {\n"
            "commandlength = c;\n"
            "} else if (serialreadpos - commandlength > DATABYTEPOSITION + 1 ) { //stx cmd len cs cs (len = 0; pos = 4)\n"
            "endread();\n"
            "continue;\n"
            "}\n"
            "else if (serialreadpos - commandlength == DATABYTEPOSITION + 1) {\n"
            "validate_serial_command();\n"
            "endread();\n"
            "continue;\n"
            "}\n"
            "}\n"
            "serialreadpos++;\n"
            "}\n",
        ],
        **{
            portcommand.name
            + "_"
            + str(portcommand.byteid): [
                "void",
                [("uint8_t*", "data"), ("uint8_t", "s")],
                portcommand.arduino_code,
            ]
            for portcommand in self.portcommands
        },
    }


def setup(self):
    setup = (
        "Serial.begin(BAUD);\n"
        "while (!Serial) {;}\n"
        "for (int i = STARTANALOG; i < ENDANALOG; i++) {\n"
        "randomSeed(analogRead(i)*random());\n"
        "}\n"
        "checkUUID();\n"
        "for (uint8_t i = 0; i < MAXFUNCTIONS; i++ ) {\n"
        "cmds[i] = 255;\n"
        "}\n"
        "ct = millis();\n"
    )
    for name, pin in self.pins.items():
        setup += "pinMode(" + name + ", " + pin.arduinoMode() + ");\n"
    for portcommand in self.portcommands:
        setup += (
            "add_command("
            + str(portcommand.byteid)
            + ", "
            + str(portcommand.sendlength)
            + ", "
            + portcommand.name
            + "_"
            + str(portcommand.byteid)
            + ");\n"
        )
    return setup


def loop(self):
    return (
        "readloop();\n"
        "ct = millis();\n"
        "if(ct-lastdata>datarate && identified){\n"
        "dataloop();\n"
        "lastdata=ct;\n"
        "}\n"
    )


def dataloop(self):
    return ""


def create(self):
    return {
        "definitions": definitions(self),
        "global_vars": global_vars(self),
        "includes": includes(self),
        "functions": functions(self),
        "setup": setup(self),
        "loop": loop(self),
        "dataloop": dataloop(self),
    }



import setuptools

setup = dict(
    name="Multi Purpose Arduino Controller",
    version="0.1",
    author="Julian Kimmig",
    author_email="julian-kimmig@gmxREADME.md.net",
    description="Controll Arduino modules",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    #py_modules=["multi_purpose_arduino_controller"],
    url="https://github.com/JulianKimmig/multi_purpose_arduino_controller",
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=["numpy","pandas","pyserial","django-bootstrap4","custom-accounts-django","django","django-glrm"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
if __name__ == "__main__":
    setuptools.setup(**setup)
